// I2Cdev library collection - MPU6050 I2C device class
// Based on InvenSense MPU-6050 register map document rev. 2.0, 5/19/2011 (RM-MPU-6000A-00)
// 8/24/2011 by Jeff Rowberg <jeff@rowberg.net>
// Updates should (hopefully) always be available at https://github.com/jrowberg/i2cdevlib
//
// Changelog:
//     ... - ongoing debug release

// NOTE: THIS IS ONLY A PARIAL RELEASE. THIS DEVICE CLASS IS CURRENTLY UNDERGOING ACTIVE
// DEVELOPMENT AND IS STILL MISSING SOME IMPORTANT FEATURES. PLEASE KEEP THIS IN MIND IF
// YOU DECIDE TO USE THIS PARTICULAR CODE FOR ANYTHING.

/* 
 Changed to C and removed unwanted functions by Alex from insideGadgets (http://www.insidegadgets.com)
 Last Modified: 17/04/2015
*/

/* ============================================
I2Cdev device library code is placed under the MIT license
Copyright (c) 2012 Jeff Rowberg

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
*/

#include "MPU6050.h"

uint8_t devAddr = MPU6050_DEFAULT_ADDRESS;

void MPU6050_dmpGetQuaternion(struct Quaternion *q, const uint8_t* packet) {
	int16_t qI[4];
	
	qI[0] = ((packet[0] << 8) + packet[1]);
	qI[1] = ((packet[4] << 8) + packet[5]);
	qI[2] = ((packet[8] << 8) + packet[9]);
	qI[3] = ((packet[12] << 8) + packet[13]);

	q->w = (float)qI[0] / 16384.0f;
	q->x = (float)qI[1] / 16384.0f;
	q->y = (float)qI[2] / 16384.0f;
	q->z = (float)qI[3] / 16384.0f;
}

void MPU6050_dmpGetGravity(struct VectorFloat *v, struct Quaternion *q) {
	v -> x = 2 * (q -> x*q -> z - q -> w*q -> y);
	v -> y = 2 * (q -> w*q -> x + q -> y*q -> z);
	v -> z = q -> w*q -> w - q -> x*q -> x - q -> y*q -> y + q -> z*q -> z;
}

void MPU6050_dmpGetYawPitchRoll(float *data, struct Quaternion *q, struct VectorFloat *gravity) {
	// yaw: (about Z axis)
	data[0] = atan2(2*q -> x*q -> y - 2*q -> w*q -> z, 2*q -> w*q -> w + 2*q -> x*q -> x - 1);
	// pitch: (nose up/down, about Y axis)
	data[1] = atan(gravity -> x / sqrt(gravity -> y*gravity -> y + gravity -> z*gravity -> z));
	// roll: (tilt left/right, about X axis)
	data[2] = atan(gravity -> y / sqrt(gravity -> x*gravity -> x + gravity -> z*gravity -> z));
	
	// Change to degrees
	data[0] = data[0] * 180/M_PI;
	data[1] = data[1] * 180/M_PI;
	data[2] = data[2] * 180/M_PI;
}


/** Power on and prepare for general usage.
 * This will activate the device and take it out of sleep mode (which must be done
 * after start-up). This function also sets both the accelerometer and the gyroscope
 * to their most sensitive settings, namely +/- 2g and +/- 250 degrees/sec, and sets
 * the clock source to use the X Gyro for reference, which is slightly better than
 * the default internal clock source.
 */
 
// The reset value is 0x00 for all registers other than the registers below.
// Register 107: 0x40.
// Register 117: 0x68.
void MPU6050_initialize(void) {
	MPU6050_setClockSource(MPU6050_CLOCK_PLL_XGYRO);
	MPU6050_setFullScaleGyroRange(MPU6050_GYRO_FS_250);
	MPU6050_setFullScaleAccelRange(MPU6050_ACCEL_FS_2);
	MPU6050_setSleepEnabled(false); // thanks to Jack Elston for pointing this one out!
}

void MPU6050_setClockSource(uint8_t source) {
	I2Cdev_writeBits(devAddr, MPU6050_RA_PWR_MGMT_1, MPU6050_PWR1_CLKSEL_BIT, MPU6050_PWR1_CLKSEL_LENGTH, source);
}

void MPU6050_setFullScaleGyroRange(uint8_t range) {
	I2Cdev_writeBits(devAddr, MPU6050_RA_GYRO_CONFIG, MPU6050_GCONFIG_FS_SEL_BIT, MPU6050_GCONFIG_FS_SEL_LENGTH, range);
}

void MPU6050_setFullScaleAccelRange(uint8_t range) {
	I2Cdev_writeBits(devAddr, MPU6050_RA_ACCEL_CONFIG, MPU6050_ACONFIG_AFS_SEL_BIT, MPU6050_ACONFIG_AFS_SEL_LENGTH, range);
}

void MPU6050_setSleepEnabled(bool enabled) {
	I2Cdev_writeBit(devAddr, MPU6050_RA_PWR_MGMT_1, MPU6050_PWR1_SLEEP_BIT, enabled);
}

void MPU6050_reset(void) {
   I2Cdev_writeBit(devAddr, MPU6050_RA_PWR_MGMT_1, MPU6050_PWR1_DEVICE_RESET_BIT, true);
}

int8_t MPU6050_getXGyroOffsetTC(void) {
	I2Cdev_readBits(devAddr, MPU6050_RA_XG_OFFS_TC, MPU6050_TC_OFFSET_BIT, MPU6050_TC_OFFSET_LENGTH, buffer);
    return buffer[0];
}

int8_t MPU6050_getYGyroOffsetTC(void) {
	 I2Cdev_readBits(devAddr, MPU6050_RA_YG_OFFS_TC, MPU6050_TC_OFFSET_BIT, MPU6050_TC_OFFSET_LENGTH, buffer);
    return buffer[0];
}

int8_t MPU6050_getZGyroOffsetTC(void) {
	I2Cdev_readBits(devAddr, MPU6050_RA_ZG_OFFS_TC, MPU6050_TC_OFFSET_BIT, MPU6050_TC_OFFSET_LENGTH, buffer);
    return buffer[0];
}

bool MPU6050_writeProgDMPConfigurationSet(const uint8_t *data, uint16_t dataSize) {
    return MPU6050_writeDMPConfigurationSet(data, dataSize, true);
}

bool MPU6050_writeProgMemoryBlock(const uint8_t *data, uint16_t dataSize, uint8_t bank, uint8_t address, bool verify) {
    return MPU6050_writeMemoryBlock(data, dataSize, bank, address, verify, true);
}

void MPU6050_setMemoryBank(uint8_t bank, bool prefetchEnabled, bool userBank) {
    bank &= 0x1F;
    if (userBank) bank |= 0x20;
    if (prefetchEnabled) bank |= 0x40;
    I2Cdev_writeByte(devAddr, MPU6050_RA_BANK_SEL, bank);
}

void MPU6050_setMemoryStartAddress(uint8_t address) {
   I2Cdev_writeByte(devAddr, MPU6050_RA_MEM_START_ADDR, address);
}

/** Verify the I2C connection.
 * Make sure the device is connected and responds as expected.
 * @return True if connection is valid, false otherwise
 */
bool MPU6050_testConnection(void) {
    return MPU6050_getDeviceID() == 0x34;
}

/** Get Device ID.
 * This register is used to verify the identity of the device (0b110100, 0x34).
 * @return Device ID (6 bits only! should be 0x34)
 * @see MPU6050_RA_WHO_AM_I
 * @see MPU6050_WHO_AM_I_BIT
 * @see MPU6050_WHO_AM_I_LENGTH
 */
uint8_t MPU6050_getDeviceID(void) {
    I2Cdev_readBits(devAddr, MPU6050_RA_WHO_AM_I, MPU6050_WHO_AM_I_BIT, MPU6050_WHO_AM_I_LENGTH, buffer);
    return buffer[0];
}

bool MPU6050_writeMemoryBlock(const uint8_t *data, uint16_t dataSize, uint8_t bank, uint8_t address, bool verify, bool useProgMem) {
    MPU6050_setMemoryBank(bank, false, false);
    MPU6050_setMemoryStartAddress(address);
    uint8_t chunkSize;
    uint8_t *verifyBuffer;
    uint8_t *progBuffer = 0;
    uint16_t i;
    uint8_t j;
    if (verify) verifyBuffer = (uint8_t *)malloc(MPU6050_DMP_MEMORY_CHUNK_SIZE);
    if (useProgMem) progBuffer = (uint8_t *)malloc(MPU6050_DMP_MEMORY_CHUNK_SIZE);
    for (i = 0; i < dataSize;) {
        // determine correct chunk size according to bank position and data size
        chunkSize = MPU6050_DMP_MEMORY_CHUNK_SIZE;

        // make sure we don't go past the data size
        if (i + chunkSize > dataSize) chunkSize = dataSize - i;

        // make sure this chunk doesn't go past the bank boundary (256 bytes)
        if (chunkSize > 256 - address) chunkSize = 256 - address;
        
        if (useProgMem) {
            // write the chunk of data as specified
            for (j = 0; j < chunkSize; j++) progBuffer[j] = pgm_read_byte(data + i + j);
        } else {
            // write the chunk of data as specified
            progBuffer = (uint8_t *)data + i;
        }

        I2Cdev_writeBytes(devAddr, MPU6050_RA_MEM_R_W, chunkSize, progBuffer);

        // verify data if needed
        if (verify && verifyBuffer) {
            MPU6050_setMemoryBank(bank, false, false);
            MPU6050_setMemoryStartAddress(address);
            I2Cdev_readBytes(devAddr, MPU6050_RA_MEM_R_W, chunkSize, verifyBuffer);
            if (memcmp(progBuffer, verifyBuffer, chunkSize) != 0) {
                free(verifyBuffer);
                if (useProgMem) free(progBuffer);
                return false; // uh oh.
            }
        }

        // increase byte index by [chunkSize]
        i += chunkSize;

        // uint8_t automatically wraps to 0 at 256
        address += chunkSize;

        // if we aren't done, update bank (if necessary) and address
        if (i < dataSize) {
            if (address == 0) bank++;
            MPU6050_setMemoryBank(bank, false, false);
            MPU6050_setMemoryStartAddress(address);
        }
    }
    if (verify) free(verifyBuffer);
    if (useProgMem) free(progBuffer);
    return true;
}


bool MPU6050_writeDMPConfigurationSet(const uint8_t *data, uint16_t dataSize, bool useProgMem) {
    uint8_t *progBuffer = 0, success, special;
    uint16_t i, j;
    if (useProgMem) {
        progBuffer = (uint8_t *)malloc(8); // assume 8-byte blocks, realloc later if necessary
    }

    // config set data is a long string of blocks with the following structure:
    // [bank] [offset] [length] [byte[0], byte[1], ..., byte[length]]
    uint8_t bank, offset, length;
    for (i = 0; i < dataSize;) {
        if (useProgMem) {
            bank = pgm_read_byte(data + i++);
            offset = pgm_read_byte(data + i++);
            length = pgm_read_byte(data + i++);
        } else {
            bank = data[i++];
            offset = data[i++];
            length = data[i++];
        }

        // write data or perform special action
        if (length > 0) {
            // regular block of data to write
            if (useProgMem) {
                if (sizeof(progBuffer) < length) progBuffer = (uint8_t *)realloc(progBuffer, length);
                for (j = 0; j < length; j++) progBuffer[j] = pgm_read_byte(data + i + j);
            } else {
                progBuffer = (uint8_t *)data + i;
            }
            success = MPU6050_writeMemoryBlock(progBuffer, length, bank, offset, true, false);
            i += length;
        } else {
            // special instruction
            // NOTE: this kind of behavior (what and when to do certain things)
            // is totally undocumented. This code is in here based on observed
            // behavior only, and exactly why (or even whether) it has to be here
            // is anybody's guess for now.
            if (useProgMem) {
                special = pgm_read_byte(data + i++);
            } else {
                special = data[i++];
            }
            if (special == 0x01) {
                // enable DMP-related interrupts
                I2Cdev_writeByte(devAddr, MPU6050_RA_INT_ENABLE, 0x32);  // single operation

                success = true;
            } else {
                // unknown special command
                success = false;
            }
        }
        
        if (!success) {
            if (useProgMem) free(progBuffer);
            return false; // uh oh
        }
    }
    if (useProgMem) free(progBuffer);
    return true;
}

void MPU6050_setIntEnabled(uint8_t enabled) {
    I2Cdev_writeByte(devAddr, MPU6050_RA_INT_ENABLE, enabled);
}

void MPU6050_setRate(uint8_t rate) {
    I2Cdev_writeByte(devAddr, MPU6050_RA_SMPLRT_DIV, rate);
}

void MPU6050_setExternalFrameSync(uint8_t sync) {
    I2Cdev_writeBits(devAddr, MPU6050_RA_CONFIG, MPU6050_CFG_EXT_SYNC_SET_BIT, MPU6050_CFG_EXT_SYNC_SET_LENGTH, sync);
}

void MPU6050_setDLPFMode(uint8_t mode) {
    I2Cdev_writeBits(devAddr, MPU6050_RA_CONFIG, MPU6050_CFG_DLPF_CFG_BIT, MPU6050_CFG_DLPF_CFG_LENGTH, mode);
}

void MPU6050_setDMPConfig1(uint8_t config) {
    I2Cdev_writeByte(devAddr, MPU6050_RA_DMP_CFG_1, config);
}

void MPU6050_setDMPConfig2(uint8_t config) {
    I2Cdev_writeByte(devAddr, MPU6050_RA_DMP_CFG_2, config);
}

void MPU6050_setOTPBankValid(bool enabled) {
    I2Cdev_writeBit(devAddr, MPU6050_RA_XG_OFFS_TC, MPU6050_TC_OTP_BNK_VLD_BIT, enabled);
}

void MPU6050_setXGyroOffsetTC(int8_t offset) {
    I2Cdev_writeBits(devAddr, MPU6050_RA_XG_OFFS_TC, MPU6050_TC_OFFSET_BIT, MPU6050_TC_OFFSET_LENGTH, offset);
}

void MPU6050_setYGyroOffsetTC(int8_t offset) {
    I2Cdev_writeBits(devAddr, MPU6050_RA_YG_OFFS_TC, MPU6050_TC_OFFSET_BIT, MPU6050_TC_OFFSET_LENGTH, offset);
}

void MPU6050_setZGyroOffsetTC(int8_t offset) {
    I2Cdev_writeBits(devAddr, MPU6050_RA_ZG_OFFS_TC, MPU6050_TC_OFFSET_BIT, MPU6050_TC_OFFSET_LENGTH, offset);
}

void MPU6050_resetFIFO(void) {
    I2Cdev_writeBit(devAddr, MPU6050_RA_USER_CTRL, MPU6050_USERCTRL_FIFO_RESET_BIT, true);
}

void MPU6050_getFIFOBytes(uint8_t *data, uint8_t length) {
   I2Cdev_readBytes(devAddr, MPU6050_RA_FIFO_R_W, length, data);
}

void MPU6050_setMotionDetectionThreshold(uint8_t threshold) {
    I2Cdev_writeByte(devAddr, MPU6050_RA_MOT_THR, threshold);
}

void MPU6050_setZeroMotionDetectionThreshold(uint8_t threshold) {
    I2Cdev_writeByte(devAddr, MPU6050_RA_ZRMOT_THR, threshold);
}

void MPU6050_setMotionDetectionDuration(uint8_t duration) {
    I2Cdev_writeByte(devAddr, MPU6050_RA_MOT_DUR, duration);
}

void MPU6050_setZeroMotionDetectionDuration(uint8_t duration) {
    I2Cdev_writeByte(devAddr, MPU6050_RA_ZRMOT_DUR, duration);
}

void MPU6050_setFIFOEnabled(bool enabled) {
    I2Cdev_writeBit(devAddr, MPU6050_RA_USER_CTRL, MPU6050_USERCTRL_FIFO_EN_BIT, enabled);
}

void MPU6050_setDMPEnabled(bool enabled) {
    I2Cdev_writeBit(devAddr, MPU6050_RA_USER_CTRL, MPU6050_USERCTRL_DMP_EN_BIT, enabled);
}

void MPU6050_resetDMP(void) {
    I2Cdev_writeBit(devAddr, MPU6050_RA_USER_CTRL, MPU6050_USERCTRL_DMP_RESET_BIT, true);
}

uint8_t MPU6050_getIntStatus(void) {
    I2Cdev_readByte(devAddr, MPU6050_RA_INT_STATUS, buffer);
    return buffer[0];
}

uint16_t MPU6050_getFIFOCount(void) {
    I2Cdev_readBytes(devAddr, MPU6050_RA_FIFO_COUNTH, 2, buffer);
    return (((uint16_t)buffer[0]) << 8) | buffer[1];
}

void MPU6050_setXGyroOffset(int16_t offset) {
    I2Cdev_writeWord(devAddr, MPU6050_RA_XG_OFFS_USRH, offset);
}

void MPU6050_setYGyroOffset(int16_t offset) {
    I2Cdev_writeWord(devAddr, MPU6050_RA_YG_OFFS_USRH, offset);
}

void MPU6050_setZGyroOffset(int16_t offset) {
    I2Cdev_writeWord(devAddr, MPU6050_RA_ZG_OFFS_USRH, offset);
}

void MPU6050_setZAccelOffset(int16_t offset) {
    I2Cdev_writeWord(devAddr, MPU6050_RA_ZA_OFFS_H, offset);
}

/** Set the I2C address of the specified slave (0-3).
 * @param num Slave number (0-3)
 * @param address New address for specified slave
 * @see getSlaveAddress()
 * @see MPU6050_RA_I2C_SLV0_ADDR
 */
void MPU6050_setSlaveAddress(uint8_t num, uint8_t address) {
    if (num > 3) return;
    I2Cdev_writeByte(devAddr, MPU6050_RA_I2C_SLV0_ADDR + num*3, address);
}

/** Set I2C Master Mode enabled status.
 * @param enabled New I2C Master Mode enabled status
 * @see getI2CMasterModeEnabled()
 * @see MPU6050_RA_USER_CTRL
 * @see MPU6050_USERCTRL_I2C_MST_EN_BIT
 */
void MPU6050_setI2CMasterModeEnabled(bool enabled) {
    I2Cdev_writeBit(devAddr, MPU6050_RA_USER_CTRL, MPU6050_USERCTRL_I2C_MST_EN_BIT, enabled);
}

/** Reset the I2C Master.
 * This bit resets the I2C Master when set to 1 while I2C_MST_EN equals 0.
 * This bit automatically clears to 0 after the reset has been triggered.
 * @see MPU6050_RA_USER_CTRL
 * @see MPU6050_USERCTRL_I2C_MST_RESET_BIT
 */
void MPU6050_resetI2CMaster(void) {
    I2Cdev_writeBit(devAddr, MPU6050_RA_USER_CTRL, MPU6050_USERCTRL_I2C_MST_RESET_BIT, true);
}

// MEM_R_W register
uint8_t MPU6050_readMemoryByte(void) {
    I2Cdev_readByte(devAddr, MPU6050_RA_MEM_R_W, buffer);
    return buffer[0];
}

// XG_OFFS_TC register
uint8_t MPU6050_getOTPBankValid() {
    I2Cdev_readBit(devAddr, MPU6050_RA_XG_OFFS_TC, MPU6050_TC_OTP_BNK_VLD_BIT, buffer);
    return buffer[0];
}

void MPU6050_readMemoryBlock(uint8_t *data, uint16_t dataSize, uint8_t bank, uint8_t address) {
    MPU6050_setMemoryBank(bank, false, false);
    MPU6050_setMemoryStartAddress(address);
    uint8_t chunkSize;
    for (uint16_t i = 0; i < dataSize;) {
        // determine correct chunk size according to bank position and data size
        chunkSize = MPU6050_DMP_MEMORY_CHUNK_SIZE;

        // make sure we don't go past the data size
        if (i + chunkSize > dataSize) chunkSize = dataSize - i;

        // make sure this chunk doesn't go past the bank boundary (256 bytes)
        if (chunkSize > 256 - address) chunkSize = 256 - address;

        // read the chunk of data as specified
        I2Cdev_readBytes(devAddr, MPU6050_RA_MEM_R_W, chunkSize, data + i);
        
        // increase byte index by [chunkSize]
        i += chunkSize;

        // uint8_t automatically wraps to 0 at 256
        address += chunkSize;

        // if we aren't done, update bank (if necessary) and address
        if (i < dataSize) {
            if (address == 0) bank++;
            MPU6050_setMemoryBank(bank, false, false);
            MPU6050_setMemoryStartAddress(address);
        }
    }
}

uint8_t MPU6050_dmpInitialize(void) {
   // reset device
   MPU6050_reset();
   _delay_ms(30); // wait after reset
	
	// disable sleep mode
   MPU6050_setSleepEnabled(false);
	
	// get MPU hardware revision
	//DEBUG_PRINTLN(F("Selecting user bank 16..."));
	MPU6050_setMemoryBank(0x10, true, true);
	//DEBUG_PRINTLN(F("Selecting memory byte 6..."));
	MPU6050_setMemoryStartAddress(0x06);
	//DEBUG_PRINTLN(F("Checking hardware revision..."));
	//uint8_t hwRevision = MPU6050_readMemoryByte();
	//DEBUG_PRINT(F("Revision @ user[16][6] = "));
	//DEBUG_PRINTLNF(hwRevision, HEX);
	//DEBUG_PRINTLN(F("Resetting memory bank selection to 0..."));
	MPU6050_setMemoryBank(0, false, false);

	// check OTP bank valid
	//DEBUG_PRINTLN(F("Reading OTP bank valid flag..."));
	//uint8_t otpValid = MPU6050_getOTPBankValid();
	//DEBUG_PRINT(F("OTP bank is "));
	//DEBUG_PRINTLN(otpValid ? F("valid!") : F("invalid!"));

	// get X/Y/Z gyro offsets
	int8_t xgOffsetTC = MPU6050_getXGyroOffsetTC();
	int8_t ygOffsetTC = MPU6050_getYGyroOffsetTC();
	int8_t zgOffsetTC = MPU6050_getZGyroOffsetTC();
	
	// setup weird slave stuff (?)
	//Setting slave 0 address to 0x7F
	MPU6050_setSlaveAddress(0, 0x7F);
	// Disabling I2C Master mode
	MPU6050_setI2CMasterModeEnabled(false);
	// Setting slave 0 address to 0x68 (self)
	MPU6050_setSlaveAddress(0, 0x68);
	// Resetting I2C Master control.
	MPU6050_resetI2CMaster();
	_delay_ms(20);

	// load DMP code into memory banks
	 if (MPU6050_writeProgMemoryBlock(dmpMemory, MPU6050_DMP_CODE_SIZE, 0, 0, true)) {
	  
	  // write DMP configuration
	  if (MPU6050_writeProgDMPConfigurationSet(dmpConfig, MPU6050_DMP_CONFIG_SIZE)) {
			// DEBUG_PRINTLN(F("Success! DMP configuration written and verified."));
			PORTB |= (1<<PB2);
			// DEBUG_PRINTLN(F("Setting clock source to Z Gyro..."));
			MPU6050_setClockSource(MPU6050_CLOCK_PLL_ZGYRO);

			// DEBUG_PRINTLN(F("Setting DMP and FIFO_OFLOW interrupts enabled..."));
			MPU6050_setIntEnabled(0x12);

			// DEBUG_PRINTLN(F("Setting sample rate to 200Hz..."));
			MPU6050_setRate(2); // 1khz / (1 + 4) = 200 Hz

			// DEBUG_PRINTLN(F("Setting external frame sync to TEMP_OUT_L[0]..."));
			MPU6050_setExternalFrameSync(MPU6050_EXT_SYNC_TEMP_OUT_L);

			// DEBUG_PRINTLN(F("Setting DLPF bandwidth to 42Hz..."));
			MPU6050_setDLPFMode(MPU6050_DLPF_BW_42);

			// DEBUG_PRINTLN(F("Setting gyro sensitivity to +/- 2000 deg/sec..."));
			MPU6050_setFullScaleGyroRange(MPU6050_GYRO_FS_2000);

			// DEBUG_PRINTLN(F("Setting DMP configuration bytes (function unknown)..."));
			MPU6050_setDMPConfig1(0x03);
			MPU6050_setDMPConfig2(0x00);

			// DEBUG_PRINTLN(F("Clearing OTP Bank flag..."));
			MPU6050_setOTPBankValid(false);

			// DEBUG_PRINTLN(F("Setting X/Y/Z gyro offset TCs to previous values..."));
			MPU6050_setXGyroOffsetTC(xgOffsetTC);
			MPU6050_setYGyroOffsetTC(ygOffsetTC);
			MPU6050_setZGyroOffsetTC(zgOffsetTC);

			// DEBUG_PRINTLN(F("Writing final memory update 1/7 (function unknown)..."));
			uint8_t dmpUpdate[16], j;
			uint16_t pos = 0;
			for (j = 0; j < 4 || j < dmpUpdate[2] + 3; j++, pos++) dmpUpdate[j] = pgm_read_byte(&dmpUpdates[pos]);
			MPU6050_writeMemoryBlock(dmpUpdate + 3, dmpUpdate[2], dmpUpdate[0], dmpUpdate[1], true, false);

			// DEBUG_PRINTLN(F("Writing final memory update 2/7 (function unknown)..."));
			for (j = 0; j < 4 || j < dmpUpdate[2] + 3; j++, pos++) dmpUpdate[j] = pgm_read_byte(&dmpUpdates[pos]);
			MPU6050_writeMemoryBlock(dmpUpdate + 3, dmpUpdate[2], dmpUpdate[0], dmpUpdate[1], true, false);

			// DEBUG_PRINTLN(F("Resetting FIFO..."));
			MPU6050_resetFIFO();

			// DEBUG_PRINTLN(F("Reading FIFO count..."));
			uint16_t fifoCount = MPU6050_getFIFOCount();
			uint8_t fifoBuffer[128];

			// DEBUG_PRINT(F("Current FIFO count="));
			// DEBUG_PRINTLN(fifoCount);
			MPU6050_getFIFOBytes(fifoBuffer, fifoCount);

			// DEBUG_PRINTLN(F("Setting motion detection threshold to 2..."));
			MPU6050_setMotionDetectionThreshold(2);

			// DEBUG_PRINTLN(F("Setting zero-motion detection threshold to 156..."));
			MPU6050_setZeroMotionDetectionThreshold(156);

			// DEBUG_PRINTLN(F("Setting motion detection duration to 80..."));
			MPU6050_setMotionDetectionDuration(80);

			// DEBUG_PRINTLN(F("Setting zero-motion detection duration to 0..."));
			MPU6050_setZeroMotionDetectionDuration(0);

			// DEBUG_PRINTLN(F("Resetting FIFO..."));
			MPU6050_resetFIFO();

			// DEBUG_PRINTLN(F("Enabling FIFO..."));
			MPU6050_setFIFOEnabled(true);

			// DEBUG_PRINTLN(F("Enabling DMP..."));
			MPU6050_setDMPEnabled(true);

			// DEBUG_PRINTLN(F("Resetting DMP..."));
			MPU6050_resetDMP();

			// // DEBUG_PRINTLN(F("Writing final memory update 3/7 (function unknown)..."));
			for (j = 0; j < 4 || j < dmpUpdate[2] + 3; j++, pos++) dmpUpdate[j] = pgm_read_byte(&dmpUpdates[pos]);
			MPU6050_writeMemoryBlock(dmpUpdate + 3, dmpUpdate[2], dmpUpdate[0], dmpUpdate[1], true, false);

			// DEBUG_PRINTLN(F("Writing final memory update 4/7 (function unknown)..."));
			for (j = 0; j < 4 || j < dmpUpdate[2] + 3; j++, pos++) dmpUpdate[j] = pgm_read_byte(&dmpUpdates[pos]);
			MPU6050_writeMemoryBlock(dmpUpdate + 3, dmpUpdate[2], dmpUpdate[0], dmpUpdate[1], true, false);

			// DEBUG_PRINTLN(F("Writing final memory update 5/7 (function unknown)..."));
			for (j = 0; j < 4 || j < dmpUpdate[2] + 3; j++, pos++) dmpUpdate[j] = pgm_read_byte(&dmpUpdates[pos]);
			MPU6050_writeMemoryBlock(dmpUpdate + 3, dmpUpdate[2], dmpUpdate[0], dmpUpdate[1], true, false);

			// DEBUG_PRINTLN(F("Waiting for FIFO count > 2..."));
			while ((fifoCount = MPU6050_getFIFOCount()) < 3);

			// DEBUG_PRINT(F("Current FIFO count="));
			// DEBUG_PRINTLN(fifoCount);
			// DEBUG_PRINTLN(F("Reading FIFO data..."));
			MPU6050_getFIFOBytes(fifoBuffer, fifoCount);

			// DEBUG_PRINTLN(F("Reading interrupt status..."));
			//uint8_t mpuIntStatus = MPU6050_getIntStatus();

			// DEBUG_PRINT(F("Current interrupt status="));
			// DEBUG_PRINTLNF(mpuIntStatus, HEX);

			// DEBUG_PRINTLN(F("Reading final memory update 6/7 (function unknown)..."));
			for (j = 0; j < 4 || j < dmpUpdate[2] + 3; j++, pos++) dmpUpdate[j] = pgm_read_byte(&dmpUpdates[pos]);
			MPU6050_readMemoryBlock(dmpUpdate + 3, dmpUpdate[2], dmpUpdate[0], dmpUpdate[1]);

			// DEBUG_PRINTLN(F("Waiting for FIFO count > 2..."));
			while ((fifoCount = MPU6050_getFIFOCount()) < 3);

			// DEBUG_PRINT(F("Current FIFO count="));
			// DEBUG_PRINTLN(fifoCount);

			// DEBUG_PRINTLN(F("Reading FIFO data..."));
			MPU6050_getFIFOBytes(fifoBuffer, fifoCount);

			// DEBUG_PRINTLN(F("Reading interrupt status..."));
			//mpuIntStatus = MPU6050_getIntStatus();

			// DEBUG_PRINT(F("Current interrupt status="));
			// DEBUG_PRINTLNF(mpuIntStatus, HEX);

			// DEBUG_PRINTLN(F("Writing final memory update 7/7 (function unknown)..."));
			for (j = 0; j < 4 || j < dmpUpdate[2] + 3; j++, pos++) dmpUpdate[j] = pgm_read_byte(&dmpUpdates[pos]);
			MPU6050_writeMemoryBlock(dmpUpdate + 3, dmpUpdate[2], dmpUpdate[0], dmpUpdate[1], true, false);

			// DEBUG_PRINTLN(F("DMP is good to go! Finally."));

			// DEBUG_PRINTLN(F("Disabling DMP (you turn it on later)..."));
			MPU6050_setDMPEnabled(false);

			// DEBUG_PRINTLN(F("Resetting FIFO and clearing INT status one last time..."));
			MPU6050_resetFIFO();
			MPU6050_getIntStatus();
		}
	}
	
	return 0; // success
}
