/*

 16MHz
 avrdude -p atmega168p -c usbtiny -U lfuse:w:0xff:m -U hfuse:w:0xdf:m -U efuse:w:0xf9:m

 
 */
 
// ATmega48A/88A/168PA/328P Pin map
//
//        21065432
//        DDDCCCCC
//        PPPPPPPP
//
//        21098765
//        33322222            
//       /--------+
//  PD3 1|        |24 PC1
//  PD4 2|        |23 PC0
//  GND 3|        |22 ADC7
//  VCC 4|        |21 GND
//  GND 5|        |20 AREF
//  VCC 6|        |19 ADC6
//  PB6 7|        |18 AVCC
//  PB7 8|        |17 PB5
//       +--------+
//        91111111
//         0123456
//
//        PPPPPPPP
//        DDDBBBBB
//        56701234

#define F_CPU 16000000 // 20 MHz

#define frontLeftPWM		OCR0A
#define frontRightPWM 	OCR0B
#define backLeftPWM 		OCR2A
#define backRightPWM		OCR2B

#define TWI_SDA_PIN PD4
#define TWI_SCL_PIN PD7
#define YAW 0
#define PITCH 1
#define ROLL 2

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <util/twi.h>
#include <stdlib.h> 
#include <math.h>
#include <string.h>
#include "setup.c"
#include "I2Cdev.c"
#include "MPU6050.c"

volatile uint8_t calCheck = 0;
volatile int16_t incSpeed = 255;

int main(void) {
	setup();
	
	float pitchCal = 0;
	float rollCal = 0;
	float yawCal = 0;
	
	// MPU control/status vars
	uint8_t mpuIntStatus;		// holds actual interrupt status byte from MPU
	uint16_t fifoCount;			// count of all bytes currently in FIFO
	uint8_t fifoBuffer[64];		// FIFO storage buffer
	uint8_t packetSize = 42;	// expected DMP packet size (default is 42 bytes)
	
	// Orientation/motion vars
	struct Quaternion q;				// [w, x, y, z]         quaternion container
	struct VectorFloat gravity;	// [x, y, z]            gravity vector
	float ypr[3];						// [yaw, pitch, roll]   yaw/pitch/roll container and gravity vector
	
	MPU6050_initialize();
	MPU6050_testConnection();
	MPU6050_dmpInitialize();

	MPU6050_setXGyroOffset(220);
	MPU6050_setYGyroOffset(76);
	MPU6050_setZGyroOffset(-85);
	MPU6050_setZAccelOffset(1788);
	MPU6050_setDMPEnabled(true);
	
	// Start calibration timer
	TCNT1 = 0;
	TCCR1B = (1<<CS12) | (1<<CS10);
	sbi(TIMSK1, TOIE1); // Enable overflow interrupt
	
	while(1) {
		//watchdog_sleep(T1S);
		
		// wait for MPU interrupt or extra packet(s) available
		//while (!mpuInterrupt && fifoCount < packetSize) {
		//}
		
		// reset interrupt flag and get INT_STATUS byte
		//uint8_t mpuInterrupt = false;
		mpuIntStatus = MPU6050_getIntStatus();
		
		// get current FIFO count
		fifoCount = MPU6050_getFIFOCount();
		
		// check for overflow (this should never happen unless our code is too inefficient)
		if ((mpuIntStatus & 0x10) || fifoCount == 1024) {
			// reset so we can continue cleanly
			MPU6050_resetFIFO();
		} 
		// otherwise, check for DMP data ready interrupt (this should happen frequently)
		else if (mpuIntStatus & 0x02) {
			// wait for correct available data length, should be a VERY short wait
			while (fifoCount < packetSize) { fifoCount = MPU6050_getFIFOCount(); }
			
			// read a packet from FIFO
			MPU6050_getFIFOBytes(fifoBuffer, packetSize);
			
			// track FIFO count here in case there is > 1 packet available
			// (this lets us immediately read more without waiting for an interrupt)
			fifoCount -= packetSize;
			
			// Get YPR
			MPU6050_dmpGetQuaternion(&q, fifoBuffer);
			MPU6050_dmpGetGravity(&gravity, &q);
			MPU6050_dmpGetYawPitchRoll(ypr, &q, &gravity);
			
			//ypr[ROLL] = ypr[ROLL] - 1;
			//ypr[PITCH] = ypr[PITCH] + 1;
			
			// Compare values
			if (calCheck == 1) {
				calCheck = 0;
				if ((ypr[PITCH] - pitchCal) < 0.2 && (ypr[PITCH] - pitchCal) > -0.2 && 
					(ypr[ROLL] - rollCal) < 0.2 && (ypr[ROLL] - rollCal) > -0.2 && 
					(ypr[YAW] - yawCal) < 0.2 && (ypr[YAW] - yawCal) > -0.2) {
					TCCR1B = 0; // Stop cal checks
					//PORTC &= ~(1<<PC3);
					calCheck = 2;
					
					TCNT1 = 0;
					TCCR1B = (1<<CS12);
				}
				else {
					//PORTC |= (1<<PC3);
				}
				
				pitchCal = ypr[PITCH];
				rollCal = ypr[ROLL];
				yawCal = ypr[YAW];
			}
			
			ypr[ROLL] -= rollCal;
			ypr[PITCH] -= pitchCal;
			
			if (calCheck == 2) {
				/*frontLeftPWM = 100;
				frontRightPWM = 100;
				backLeftPWM = 100;
				backRightPWM = 100;*/
				
				frontLeftPWM = incSpeed;
				frontRightPWM = incSpeed;
				backLeftPWM = incSpeed;
				backRightPWM = incSpeed;
				
				//frontLeftPWM -= 3;
				//backLeftPWM -= 3;
				//backRightPWM -= 3;
				
				int16_t maxspeed = 40;
				float dividerpos = 0.5;
				float dividerneg = -0.5;
				
				// Front right
				if (ypr[ROLL] > 0.5 && ypr[PITCH] > 0.5) {
					float FBAmount = ypr[ROLL] / (float) dividerpos;
					float LRAmount = ypr[PITCH] / (float) dividerpos;
					if (frontRightPWM - FBAmount <= maxspeed || frontRightPWM - LRAmount <= maxspeed) { // Max on
						frontRightPWM = maxspeed;
					}
					else {
						frontRightPWM -= FBAmount - LRAmount;
					}
					
					if (backLeftPWM + FBAmount >= 255 || backLeftPWM + LRAmount >= 255) { // Max off
						backLeftPWM = 255;
					}
					else {
						backLeftPWM += (FBAmount + LRAmount) / 2;
					}
				}
				
				// Front left
				if (ypr[ROLL] > 0.5 && ypr[PITCH] < -0.5) {
					float FBAmount = ypr[ROLL] / (float) dividerpos;
					float LRAmount = ypr[PITCH] / (float) dividerneg;
					
					if (frontLeftPWM - FBAmount <= maxspeed || frontLeftPWM - LRAmount <= maxspeed) { // Max on
						frontLeftPWM = maxspeed;
					}
					else {
						frontLeftPWM -= FBAmount - LRAmount;
					}
					
					if (backRightPWM + FBAmount >= 255 || backRightPWM + LRAmount >= 255) { // Max off
						backRightPWM = 255;
					}
					else {
						backRightPWM += (FBAmount + LRAmount) / 2;
					}
				}
				
				// Back right
				if (ypr[ROLL] < -0.5 && ypr[PITCH] > 0.5) {			
					float FBAmount = ypr[ROLL] / (float) dividerneg;
					float LRAmount = ypr[PITCH] / (float) dividerpos;
					
					if (backRightPWM - FBAmount <= maxspeed || backRightPWM - LRAmount <= maxspeed) { // Max on
						backRightPWM = maxspeed;
					}
					else {
						backRightPWM -= FBAmount - LRAmount;
					}
					
					if (frontLeftPWM + FBAmount >= 255 || frontLeftPWM + LRAmount >= 255) { // Max off
						backRightPWM = 255;
					}
					else {
						frontLeftPWM += (FBAmount + LRAmount) / 2;
					}
					
				}
				
				// Back left
				if (ypr[ROLL] < -0.5 && ypr[PITCH] < -0.5) {
					float FBAmount = ypr[ROLL] / (float) dividerneg;
					float LRAmount = ypr[PITCH] / (float) dividerneg;
					
					if (backLeftPWM - FBAmount <= maxspeed || backLeftPWM - LRAmount <= maxspeed) { // Max on
						backLeftPWM = maxspeed;
					}
					else {
						backLeftPWM -= FBAmount - LRAmount;
					}
					
					if (frontRightPWM + FBAmount >= 255 || frontRightPWM + LRAmount >= 255) { // Max off
						frontRightPWM = 255;
					}
					else {
						frontRightPWM += (FBAmount + LRAmount) / 2;
					}
				}
				
			}
			
			//if (startCal < 500) {
			//	startCal++;
				//rollCal = ypr[ROLL];
				//pitchCal = ypr[PITCH];
			//}
			
			if (calCheck == 2) {
				if (ypr[ROLL] > -1 && ypr[ROLL] < 1 && ypr[PITCH] > -1 && ypr[PITCH] < 1) {
					PORTC |= (1<<PC3);
				}
				else {
					PORTC &= ~(1<<PC3);
				}
			}
			
		}
	}
}

ISR(WDT_vect) { }

ISR(TIMER1_OVF_vect) {
	if (calCheck != 2) { 
		calCheck = 1;
	}
	else {
		//if (incSpeed > 190) {
		if (incSpeed > 90) {
			incSpeed -= 10;
		}
	}
}
